# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cons.ipynb.

# %% auto 0
__all__ = ['U1', 'U2', 'DOT', 'NIL', 'TILDE', 'ALT', 'CLS', 'GET', 'RAW', 'ATTR', 'DUCK', 'LOAD', 'STRS', 'TYPE', 'TYPES',
           'DTYPE', 'VALUE', 'BASES', 'CHECK', 'GUARD', 'ASSET', 'AREPR', 'ATTRS', 'PREFIX', 'SUFFIX', 'RETURN',
           'DUCKED', 'LOADER', 'MODULE', 'RETHAS', 'RETSELF', 'DEFAULT', 'CLSNAME', 'DELAYED', 'DUCKSPEC', 'DROPNONE',
           'GETPROPS', 'CHECKOBJ', 'GUARDOBJ', 'IMPORTED', 'IS_MODULE', 'DECORATED', 'NAMESPACE', 'LOADER_STATE',
           'KEYS', 'VALUES', 'NDIM', 'SHAPE', 'ALL_ATTRS', 'DUCK_SPEC_ATTRS', 'DUCK_ATTRS', 'DUCKED_ATTRS',
           'DUCK_MODULE_ATTRS', 'IPYTHON_ATTRIBUTES', 'PROTOCOL_META_ATTRS', 'ANNOYING_ATTRS', 'MODULE_TYPE_ATTRS',
           'ASSIGNED', 'ATTRIBUTE_CLASS_VARS', 'PROTOCOL_METHOD_ORDER']

# %% ../nbs/00_cons.ipynb 6
#| export


# %% ../nbs/00_cons.ipynb 8
#| export


# %% ../nbs/00_cons.ipynb 10
#| export

# %% ../nbs/00_cons.ipynb 12
#| export

# %% ../nbs/00_cons.ipynb 14
#| export

# %% ../nbs/00_cons.ipynb 17
U1 = '_'
U2 = '__'

DOT = '.'
NIL = ''

TILDE = '~'

# %% ../nbs/00_cons.ipynb 19
ALT = 'alt'
CLS = 'cls'
GET = 'get'
RAW = 'raw'

ATTR = 'attr'
DUCK = 'duck' 
LOAD = 'load'
STRS = 'strs'
TYPE = 'type' 


TYPES = 'types'
DTYPE = 'dtype'
VALUE = 'value'

BASES = 'bases' 
CHECK = 'check'
GUARD = 'guard'

ASSET = 'asset'
AREPR = 'arepr'
ATTRS = 'attrs'

PREFIX = 'prefix'
SUFFIX = 'suffix'
 
 
RETURN = 'return'
DUCKED = 'ducked'
LOADER = 'loader'
MODULE = 'module'
RETHAS = 'rethas' 
RETSELF = 'retself'

DEFAULT = 'default'
CLSNAME = 'clsname' 
DELAYED = 'delayed'

DUCKSPEC = 'duckspec'
DROPNONE = 'dropnone'
GETPROPS = 'getprops'
CHECKOBJ = 'checkobj'
GUARDOBJ = 'guardobj'
IMPORTED = 'imported'

IS_MODULE = 'is_module'
DECORATED = 'decorated'
NAMESPACE = 'namespace'

LOADER_STATE = 'loader_state'

# %% ../nbs/00_cons.ipynb 20
KEYS = 'keys'
VALUES = 'values'
NDIM = 'ndim'
SHAPE = 'shape'

# %% ../nbs/00_cons.ipynb 22
_AREPR, _ASSET, _IMPORTED, _MODULE = f'{U1}{AREPR}', f'{U1}{ASSET}', f'{U1}{IMPORTED}', f'{U1}{MODULE}'
_DUCK = f'{U1}{DUCK}'
_DECORATED = f'{U1}{DECORATED}'

# %% ../nbs/00_cons.ipynb 24
_DEFAULT_ = '_default_'

# %% ../nbs/00_cons.ipynb 26
__K__ = '__k__'
__V__ = '__v__'
__N__ = '__n__'

__KVTYPES__ = '__kvtypes__'
__RETANNOT__ = '__retannot__'

# %% ../nbs/00_cons.ipynb 27
__DOC__ = '__doc__'
__NEW__ = '__new__'

__CALL__ = '__call__'
__INIT__ = '__init__'
__HASH__ = '__hash__'
__NAME__ = '__name__'
__SPEC__ = '__spec__'


__HUSH__ = '__hush__'
__LOAD__ = '__load__'
__DUCK__ = '__duck__'
__ALTS__ = '__alts__'

__BASES__ = '__bases__'
__PROPS__ = '__props__'
__ATTRS__ = '__attrs__'


__MODULE__ = '__module__'
__QUALNAME__ = '__qualname__'
__SIGNATURE__ = '__signature__'

__DUCKTYPE__ = '__ducktype__'
__EXPECTED__= '__expected__'
__REQUIRED__= '__required__'
__ALLEQUAL__= '__allequal__'
__EXTRAFNS__= '__extrafns__'
__USEALIAS__= '__usealias__'


__DUCKTYPES__ = '__ducktypes__'
__DUCKMODULE__ = '__duckmodule__'
__ANNOTATIONS__ = '__annotations__'

# %% ../nbs/00_cons.ipynb 29
ALL_ATTRS = {TYPE, DTYPE, TYPES}
'''Attributes of which all values in the iterable are to be checked against the type guards.''';

# %% ../nbs/00_cons.ipynb 31
DUCK_SPEC_ATTRS = (ASSET, IS_MODULE, DUCKED, DEFAULT, BASES, ATTRS, DELAYED)

# %% ../nbs/00_cons.ipynb 32
DUCK_ATTRS = (__DUCKTYPE__, __EXPECTED__, __REQUIRED__, __ALLEQUAL__, __EXTRAFNS__, __USEALIAS__)
'''Attributes that ABCDuck is expected to have''';

# %% ../nbs/00_cons.ipynb 33
DUCKED_ATTRS = (DUCK, _DUCK, __DUCK__, __DUCKTYPES__)
'''Attributes that an object _might_ have if the object maybe is a duck-type.''';

# %% ../nbs/00_cons.ipynb 34
DUCK_MODULE_ATTRS = (__DUCKMODULE__, __ALTS__, __HUSH__)

# %% ../nbs/00_cons.ipynb 36
IPYTHON_ATTRIBUTES = {
    '_ipython_canary_method_should_not_exist_', 
    '_ipython_display_',
    '_repr_mimebundle_',
    '_repr_html_',
    '_repr_markdown_',
    '_repr_svg_',
    '_repr_png_',
    '_repr_pdf_',
    '_repr_jpeg_',
    '_repr_latex_',
    '_repr_json_',
    '_repr_javascript_'
}
'''Attributes that `ipython` might invoke on an object when used in a notebook or console environment.''';

PROTOCOL_META_ATTRS = {'_is_protocol', '_is_runtime_protocol'}
'''Attributes that are invoked when `isinstance` is called on a ProtocolMeta subclass.''';

ANNOYING_ATTRS = IPYTHON_ATTRIBUTES | PROTOCOL_META_ATTRS
'''Attributes that are annoying to deal with when introspecting an object and should be ignored.''';

MODULE_TYPE_ATTRS = set({'__name__', '__doc__', '__loader__', '__spec__'})
'''Attributes that are required to represent a ModuleType''';

# %% ../nbs/00_cons.ipynb 37
ASSIGNED = (__MODULE__, __DOC__, __ANNOTATIONS__)

ATTRIBUTE_CLASS_VARS = (ATTR, CHECK, GUARD, CHECKOBJ, GUARDOBJ, RETSELF, DEFAULT)

# %% ../nbs/00_cons.ipynb 38
PROTOCOL_METHOD_ORDER = ('instcheck', 'decorator', 'generator', 'typeguard')
'''Options for the overloaded `ProtoType.__call__` method.''';



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_base.ipynb.

# %% auto 0
__all__ = ['GenericDuck', 'Duck', 'DuckModule']

# %% ../nbs/10_base.ipynb 6
import sys, copy

# %% ../nbs/10_base.ipynb 8
from types import ModuleType
from typing import Protocol, Generic

# %% ../nbs/10_base.ipynb 10
#| export

# %% ../nbs/10_base.ipynb 12
#| export

# %% ../nbs/10_base.ipynb 14
from .cons import DUCK_ATTRS
from .type import P, T, Ts
from .meta import DuckModuleMeta, ABCDuck, ABCParam

# %% ../nbs/10_base.ipynb 17
class GenericParam(Generic[*Ts, P], metaclass=ABCParam):
    ...

# %% ../nbs/10_base.ipynb 19
class GenericDuck(Generic[*Ts, P], metaclass=ABCDuck): 
    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        # print('base.__metahook__', cls, mcls, name, bases, namespace, **kwargs)
        return cls

# %% ../nbs/10_base.ipynb 22
class Duck(Protocol[*Ts, P], metaclass=ABCDuck):
    '''A class representing a duck-type.'''
    def __new__(cls, *args, **kwargs):
        # print('base.__new__', cls, args, kwargs)
        super().__new__(*args, **kwargs)
        
    def __init__(cls, *args, **kwargs):
        # print('base.__init__', cls, args, kwargs)
        super().__init__(*args, **kwargs)

    @classmethod
    def __metahook__(cls, mcls, name, bases, namespace, **kwargs):
        # print('base.__metahook__', cls, mcls, name, bases, namespace, **kwargs)
        return cls
        
    def __init_subclass__(cls, *args, **kwargs):
        # print('base.__init_subclass__', cls, args, kwargs)
        super().__init_subclass__(*args, **kwargs)
        for attr in DUCK_ATTRS:
            if (aval := getattr(cls, attr, None)) is not None:
                setattr(cls, attr, copy.copy(aval))

# %% ../nbs/10_base.ipynb 30
class DuckModule(ModuleType, metaclass=DuckModuleMeta):
    ...



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_grds.ipynb.

# %% auto 0
__all__ = ['OriginTypes', 'OriginGuard', 'hashash', 'ishashable', 'hasorigin', 'isspecial', 'isnotspecial', 'allnotspecial',
           'getnotspecials', 'notspecialsorigins', 'isorigin', 'iscls', 'isstr', 'notcls', 'isnone', 'isbool', 'isdict',
           'isiter', 'isnan', 'istuple', 'notnone', 'isargs', 'istyping', 'isnotiterstr', 'isclstype', 'hasduck',
           'hasload', 'istype', 'isallsame', 'isalias']

# %% ../nbs/03_grds.ipynb 6
import io, math, inspect
from operator import eq, is_
from functools import wraps
from contextlib import redirect_stderr

# %% ../nbs/03_grds.ipynb 8
from types import FunctionType
from typing import (
    Any, Union, Callable, Iterable, TypeAlias, TypeGuard, get_args, get_origin, 
    Generic, Protocol, _SpecialForm, )

# %% ../nbs/03_grds.ipynb 10
#| export

# %% ../nbs/03_grds.ipynb 12
#| export

# %% ../nbs/03_grds.ipynb 14
from .cons import U1, U2, DUCK, LOAD, __HASH__, ALL_ATTRS
from .cons import NIL, NDIM, SHAPE, KEYS, VALUES, __K__, __V__, __N__, __RETANNOT__, __KVTYPES__
from .type import T, O
from .prot import Hashable, NonStrBytesIterable, Origin, Special, NotSpecial
from .prot import DuckSpecProtocol, DuckModuleProtocol, DuckTypeProtocol

# %% ../nbs/03_grds.ipynb 17
def hashash(o: object) -> TypeGuard[Hashable]:
    '''Check if the given object is hashable.'''
    return hasattr(o, __HASH__)

@wraps(hashash)
def ishashable(o: object) -> TypeGuard[Hashable]:
    return hashash(o)

def hasorigin(t: type) -> TypeGuard[Origin]:
    return bool(get_origin(t)) and hashash(t)

def isspecial(t: type) -> TypeGuard[Special]:
    return hasorigin(t) and isinstance(t, _SpecialForm) or t in (Generic, Protocol)

def isnotspecial(t: type) -> TypeGuard[NotSpecial]:
    return hasorigin(t) and not isspecial(t)

# %% ../nbs/03_grds.ipynb 19
def allnotspecial() -> tuple[NotSpecial, ...]:
    import typing
    return tuple(t for attr in dir(typing) if isnotspecial(t := getattr(typing, attr)))

def getnotspecials() -> Union[tuple[NotSpecial, ...]]:
    types: tuple[type, ...] = allnotspecial()
    return Union[types]

def notspecialsorigins() -> set[type]:
    return set([get_origin(o) for o in allnotspecial()])

# %% ../nbs/03_grds.ipynb 20
OriginTypes = Union[allnotspecial()]
OriginGuard: TypeAlias = TypeGuard[OriginTypes]

# %% ../nbs/03_grds.ipynb 21
def isorigin(t: type) -> OriginGuard:
    '''Check if the given type is one of the origin types defined in `OriginTypes`.'''
    return get_origin(t) in notspecialsorigins()

def ishashable(o: object) -> TypeGuard[Hashable]:
    '''Check if the given object is hashable.'''
    return hashash(o)

# %% ../nbs/03_grds.ipynb 23
def iscls(x) -> TypeGuard[type]:
    '''Check if `x` is a class.'''
    return inspect.isclass(x)

def isstr(x) -> TypeGuard[str]:
    '''Check if `x` is an `str`.'''
    return isinstance(x, str)

def notcls(x) -> TypeGuard[T]:
    '''Check if `x` is not a class.'''
    return not iscls(x)

def isnone(x) -> TypeGuard[None]:
    '''Check if `x` is `None`'''
    return x is None

def isbool(x) -> TypeGuard[bool]:
    '''Check if `x` is a `bool`'''
    return isinstance(x, bool)

def isdict(x) -> TypeGuard[dict]:
    '''Check if `x` is a `dict`'''
    return isinstance(x, dict)

def isiter(o: object) -> TypeGuard[Iterable]: 
    '''Check if the given object is an iterable.'''
    return isinstance(o, Iterable)

def isnan(o: object) -> TypeGuard[math.nan]:
    '''Check if the given object is `math.nan`.'''
    try: return math.isnan(o)
    except: return False

def _istype(x) -> TypeGuard[type]:
    '''Check if `x` is a `type`'''
    return isinstance(x, type)

def istuple(x) -> TypeGuard[tuple]:
    '''Check if `x` is a `tuple`'''
    return isinstance(x, tuple)

def notnone(x) -> TypeGuard[Any]:
    '''Check if `x` is not `None`'''
    return not isnone(x)

def isargs(o: object, t: T) -> TypeGuard[T]:
    '''Check if the given object is equal to or contained within the arguments of the type `t`.'''
    try: 
        args = get_args(t)
        # NOTE: len(args) and o == args makes sure empty tuple() is not True
        return any(((len(args) and o == args), o in args, o == t))
    except: return False

def istyping(o: object, t: T) -> TypeGuard[T]:
    '''Check if the given object is a typing construct that matches type `t`.'''
    if not isorigin(t): return False
    try: return isargs(o, t)
    except: return False
    
def isnotiterstr(o: object) -> TypeGuard[NonStrBytesIterable]:
    '''Check if the given object is an iterable but not a string or bytes.'''
    return isiter(o) and not isinstance(o, (str, bytes))

def isclstype(o: object) -> TypeGuard[type]:
    '''Check if the given object is a class type.'''
    return iscls(o) or isinstance(o, type)

# %% ../nbs/03_grds.ipynb 24
def hasduck(o: object) -> TypeGuard[T]:
    '''Check if the given object has the attribute `duck`.'''
    return hasattr(o, DUCK)

def hasload(o: object) -> TypeGuard[T]:
    '''Check if the given object has the attribute `load` and it is callable.'''
    return callable(getattr(o, LOAD, None))

# %% ../nbs/03_grds.ipynb 26
def istype(o: object, t: T, guards: Iterable[FunctionType] = tuple(), dropfn: Iterable[FunctionType] = tuple()) -> TypeGuard[T]:
    '''Check if the given object is of type `t`, using additional guard functions.

    Parameters
    ----------
    o : object
        The object to check.
        
    t : T
        The type to check against.
        
    guards : IterFunc, optional
        Additional guard functions to use in the check.
        
    dropfn : IterFunc, optional
        Functions to exclude from the check.

    Returns
    -------
    Guard
        True if the object is of type `t` or satisfies any of the guard functions, otherwise False.
        
    Notes
    -----
    The `guards` and `dropfn` parameters are used to filter the set of functions used in the check.
    
    The mandatory guard functions are `isinstance`, `issubclass`, `typs.istyping`, `typs.isargs`, 
        `operator.eq`, and `operator.is_`.
    '''
    funcs = (isinstance, issubclass, istyping, isargs, eq, is_,) + tuple(guards)
    funcs = (f for f in funcs if f not in dropfn)
    with redirect_stderr(io.StringIO()):
        for comp in funcs:
            try:
                if comp(o, t):
                    return True
            except: 
                continue
            
        if isnan(t): # checknan(o, t) ==> isnan(t) and isnan(o)
            if isnan(o): 
                return True
    # print(f'istype({o}, {t}), False')
    return False

# %% ../nbs/03_grds.ipynb 28
def isallsame(it: Iterable[T], dtype: T) -> TypeGuard[Iterable[T]]:
    '''Check if all elements in an iterable are of the same specified type.

    Parameters
    ----------
    it : IterType
        The iterable to check.
        
    dtype : T
        The type against which to check the elements of the iterable.

    Returns
    -------
    GuardIterType
        True if all elements are of the specified type, otherwise False.
    '''
    if not isiter(it): 
        return istype(it, dtype)
    
    for el in it:
        if isnotiterstr(el):
            # el is also iterable (but not like 'abc'), so recurse over the sub-iterable elements
            try:
                if not isallsame(el, dtype): 
                    return False
            except TypeError: 
                return False
            
        elif not istype(el, dtype):
            return False
    return True

# %% ../nbs/03_grds.ipynb 30
def isalias(
    obj: T,       # object to check
    types: tuple[type, ...], # types to check
    guards: Iterable[FunctionType] = tuple(),
    attrs: dict = dict(),
    hasattrs: Iterable[str] = tuple(),  # attributes to check
    allattrs: Iterable[str] = ALL_ATTRS, # attributes to check that values are all the same if object is iterable
    mapattrs: dict[str, Callable[[O, T], TypeGuard[T]]] = dict(),
) -> TypeGuard[Iterable[T]]:
    '''Check if an object matches a set of types and specific attributes.

    Parameters
    ----------
    obj : T
        The object to check.
        
    types : Types
        A tuple of types to compare against the object.
        
    guards : IterFunc, optional
        Iterable of guard functions.
        
    attrs : dict, optional
        Dictionary of attributes and their expected values to check in the object.
        
    hasattrs : IterStr, optional
        Iterable of attribute names that the object must have.
        
    allattrs : IterStr, optional
        Iterable of attributes to check if their values are all the same in an iterable object.
        
    mapattrs : AttrGuards, optional
        Mapping of attribute names to guard functions for additional checks.

    Returns
    -------
    GuardIterT
        Returns True if the object matches the specified conditions, False otherwise.
    '''
    # print('_____isalias_____')
    # print(obj, types, guards, attrs, hasattrs, allattrs, mapattrs)
    hasattrs = tuple(hasattrs) + tuple(k for k, v in attrs.items() if k not in allattrs)
    if not all(hasattr(obj, attr) for attr in hasattrs): return False
    
    
    found = any(istype(obj, t, guards) for t in types)
    if not found: return False
    
    for attr, val in attrs.items():
        if not (getattr(obj, attr, None) == val):
            if attr in hasattrs and attr not in (*allattrs, *mapattrs): return False
            # ...
        
        if attr in allattrs and not (isnotiterstr(obj) and isallsame(obj, val)): 
            return False
            # continue
        
        if attr in mapattrs and not (mapattrs[attr](obj, val)):            
            return False
            # continue
        
    return True

# %% ../nbs/03_grds.ipynb 32
def isduck(item) -> TypeGuard[Union['ABCDuck', DuckTypeProtocol]]:
    try: from quac.meta import ABCDuck as Duckie
    except: Duckie = DuckTypeProtocol
    return isinstance(item, (DuckTypeProtocol, Duckie, ))

def isduckspec(item) -> TypeGuard[Union['DuckSpec', DuckSpecProtocol]]:
    try: from quac.spec import DuckSpec as Spec
    except: Spec = DuckSpecProtocol
    return isinstance(item, (DuckSpecProtocol, Spec, ))

def isduckmodule(item) -> TypeGuard[DuckModuleProtocol]:
    try: return issubclass(item, (DuckModuleProtocol, ))
    except: return False

# %% ../nbs/03_grds.ipynb 35
def _islen(cls):
    n = getattr(cls, __N__, 0)
    def islen(obj: O, val: int = n) -> TypeGuard[cls]: 
        return len(obj) == val
    islen.__name__ = islen.__qualname__ = f'islen{n}'
    return islen

# %% ../nbs/03_grds.ipynb 37
def _istkey(cls, typek: O = None):
    '''TypeGuard to check if an object is a dict with keys of a specific type'''
    k = getattr(cls, __K__, typek)
    def istkey(obj: T, key: O = k) -> TypeGuard[cls]: 
        if k is None: return isdict(obj)
        return isdict(obj) and all(isinstance(k, (key, )) for k in obj.keys())
    istkey.__name__ = istkey.__qualname__ = f'istkey{k}'
    return istkey

def _istval(cls, typev: O = None):
    '''TypeGuard to check if an object is a dict with values of a specific type'''
    v = getattr(cls, __V__, typev)
    def istkey(obj: T, val: O = v) -> TypeGuard[cls]: 
        if v is None: return isdict(obj) 
        return isdict(obj) and all(isinstance(v, (val, )) for v in obj.values())
    
    istkey.__name__ = istkey.__qualname__ = f'istval{v}'
    return istkey

def _iskvdict(cls):
    (tk, tv) = getattr(cls, __KVTYPES__, (None, None))
    k = getattr(cls, __K__, tk)
    v = getattr(cls, __V__, tv)
    istkeys = _istkey(cls, k)
    istvals = _istval(cls, v)
    def iskvdict(obj: O, val: tuple = (k, v)) -> TypeGuard[cls]: 
        return isdict(obj) and istkeys(obj) and istvals(obj)
    iskvdict.__name__ = iskvdict.__qualname__ = f'is_{k}_{v}_dict'
    return iskvdict



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_prot.ipynb.

# %% auto 0
__all__ = ['ProtocolMeta', 'GuardProtocol', 'NotProtocol', 'OptProtocol', 'Hashable', 'Origin', 'Special', 'NotSpecial',
           'NonStrBytesIterable', 'HasAnyAttributeProtocol', 'HasAllAttributeProtocol', 'DuckedProtocol',
           'DuckSpecProtocol', 'DuckModuleProtocol', 'DuckTypeProtocol', 'AnnDataProtocol', 'NDArrayProtocol',
           'DataFrameProtocol']

# %% ../nbs/02_prot.ipynb 6
#| export


# %% ../nbs/02_prot.ipynb 8
from types import NoneType
from typing import (Any, Self, TypeGuard, Protocol, Iterable, _SpecialForm, _ProtocolMeta, Generic, get_origin,)

# %% ../nbs/02_prot.ipynb 10
#| export

# %% ../nbs/02_prot.ipynb 12
#| export

# %% ../nbs/02_prot.ipynb 14
from .cons import U1, U2, DUCK_ATTRS, DUCKED_ATTRS, DUCK_SPEC_ATTRS, __HASH__, __ATTRS__
from .type import T

# %% ../nbs/02_prot.ipynb 16
class ProtocolMeta(_ProtocolMeta):
    def __instancecheck__(cls, __instance) -> TypeGuard[T]:
        return cls.__guard__(__instance)
    
class GuardProtocol(Protocol[T], metaclass=ProtocolMeta):
    ...

# %% ../nbs/02_prot.ipynb 18
class NotProtocol(Protocol[T]):
    '''A protocol representing a not being a type'''
    def __instancecheck__(cls: type[T], inst: Any) -> TypeGuard[type[T]]: 
        return not isinstance(inst, cls)
    
class OptProtocol(Protocol[T]):
    '''A protocol representing an optional type'''
    def __instancecheck__(cls: type[T], inst: Any) -> TypeGuard[type[T]]: 
        return isinstance(inst, (cls, NoneType))

# %% ../nbs/02_prot.ipynb 20
class Hashable(Protocol):
    '''A protocol representing a hashable type'''
    def __hash__(self: Self) -> int: ...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['Hashable']:
        return hasattr(__instance, __HASH__)

# %% ../nbs/02_prot.ipynb 22
class Origin(Hashable, Protocol):
    '''A protocol representing origin typing types'''
    def __hash__(self: Self) -> int:...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['Origin']:
        return hasattr(__instance, __HASH__) and get_origin(__instance)

# %% ../nbs/02_prot.ipynb 24
class Special(Origin, Protocol):
    '''A protocol representing special typing types'''
    def __hash__(self: Self) -> int:...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['Special']:
        origin = super().__instancecheck__(__instance)
        special = isinstance(__instance, _SpecialForm) or __instance in (Generic, Protocol)
        return (origin and special)
    
class NotSpecial(Origin, Protocol):
    '''A protocol representing not special typing types'''
    def __hash__(self: Self) -> int: ...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['NotSpecial']:
        origin = super().__instancecheck__(__instance)
        special = isinstance(__instance, _SpecialForm) or __instance in (Generic, Protocol)
        return (origin and not special)

# %% ../nbs/02_prot.ipynb 26
class NonStrBytesIterable(Protocol):
    '''A protocol representing non-string, non-bytes iterable objects.'''
    def __iter__(self: Self) -> Iterable: ...
    def __instancecheck__(self: Self, __instance) -> TypeGuard['NonStrBytesIterable']:
        return isinstance(__instance, Iterable) and not isinstance(__instance, (str, bytes))

# %% ../nbs/02_prot.ipynb 28
class HasAnyAttributeProtocol(GuardProtocol[T]):
    '''A protocol representing to check if an object has any attribute in `__attrs__`.'''
    __attrs__: tuple[str, ...] = ()
    @classmethod
    def __guard__(self: Self, ins) -> TypeGuard[T]:
        attrs = getattr(self, __ATTRS__, ())
        if not any(hasattr(ins, attr) for attr in attrs): return False
        return True
    
    
class HasAllAttributeProtocol(GuardProtocol[T]):
    '''A protocol representing to check if an object has all attributes in `__attrs__`.'''
    __attrs__: tuple[str, ...] = ()
    @classmethod
    def __guard__(self: Self, ins) -> TypeGuard[T]:
        attrs = getattr(self, __ATTRS__, ())
        if not all(hasattr(ins, attr) for attr in attrs): return False
        return True

# %% ../nbs/02_prot.ipynb 30
class DuckedProtocol(HasAnyAttributeProtocol[T]):
    '''A protocol representing a something that might be like a duck'''
    __attrs__: tuple[str, ...] = DUCKED_ATTRS
    
class DuckSpecProtocol(HasAllAttributeProtocol[T]):
    '''A protocol representing a duckspec'''
    __attrs__: tuple[str, ...] = DUCK_SPEC_ATTRS
    
class DuckModuleProtocol(HasAllAttributeProtocol[T]):
    '''A protocol representing a duck module'''
    __attrs__: tuple[str, ...]
    
class DuckTypeProtocol(HasAllAttributeProtocol[T]):
    '''A protocol representing a duck-type'''
    __attrs__: tuple[str, ...] = DUCK_ATTRS

# %% ../nbs/02_prot.ipynb 33
class AnnDataProtocol(HasAllAttributeProtocol[T]):
    '''A protocol representing a `anndata.AnnData`'''
    __attrs__: tuple[str, ...] = ('X', 'obs', 'var', 'obsm', 'varm', 'uns', 'layers',)

# %% ../nbs/02_prot.ipynb 35
class NDArrayProtocol(HasAllAttributeProtocol[T]):
    '''A protocol representing a `numpy.ndarray`'''
    __attrs__: tuple[str, ...] = ('T', 'ndim', 'shape', )

# %% ../nbs/02_prot.ipynb 37
class DataFrameProtocol(HasAllAttributeProtocol[T]):
    '''A protocol representing a `pandas.DataFrame`'''
    __attrs__: tuple[str, ...] = (
        'T', 'describe', 'groupby', 'head', 'info', 'mean', 'median', 'nunique', 'quantile', 'shape', 'tail', 'isin',
        'keys', 'where', 'values', 'columns',
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_enum.ipynb.

# %% auto 0
__all__ = ['Species', 'LoaderState']

# %% ../nbs/04_enum.ipynb 6
import io
from enum import StrEnum, auto
from contextlib import redirect_stdout, redirect_stderr
from importlib.machinery import ModuleSpec

# %% ../nbs/04_enum.ipynb 8
from types import ModuleType
from typing import (Any, Union, Optional, Iterable,)

# %% ../nbs/04_enum.ipynb 10
#| export

# %% ../nbs/04_enum.ipynb 12
#| export

# %% ../nbs/04_enum.ipynb 14
from .cons import NIL, VALUE, LOADER_STATE, _DEFAULT_, __SPEC__

# %% ../nbs/04_enum.ipynb 17
class Species(StrEnum):
    '''Cheeky enum to represent the "species" (i.e. flavor) of a duck-typed protocol.'''
    REG = auto()
    NOT = auto()
    OPT = auto()
    
    @classmethod
    def contains(cls, item: Any, items: Iterable, kind: 'Species' = 'reg') -> bool:
        kind = cls(kind)
        match kind:
            case cls.REG: return item in items
            case cls.NOT: return item not in items
            case cls.OPT: return True
            case _: return False

# %% ../nbs/04_enum.ipynb 23
class LoaderState(StrEnum):
    '''The state of the module loader.

    Attributes
    ----------
    NONE : auto
        Indicates no state.
    EXECED : auto
        Indicates the module has been executed.
    DUCKED : auto
        Indicates the module has been duck-typed.
    UNKNOWN : auto
        Indicates an unknown state.

    Methods
    -------
    _missing_(value: object) -> 'LoaderState'
        Provides a default for missing values.
    spec(module: ModuleType) -> Optional[ModuleSpec]
        Retrieves the module spec.
    get(module: ModuleType) -> 'LoaderState'
        Retrieves the loader state for a module.
    set(module: ModuleType, state: Optional[Union['LoaderState', str]] = None) -> ModuleType
        Sets the loader state for a module.
    exec(module: ModuleType) -> ModuleType
        Executes the module and sets its state to DUCKED.
    '''
    NONE = auto()
    EXECED = auto()
    DUCKED = auto()
    UNKNOWN = auto()
    
    @classmethod
    def _missing_(cls, value: object) -> 'LoaderState':
        if value is None: return cls.NONE
        return cls.UNKNOWN 
                
    @classmethod
    def spec(cls, module: ModuleType) -> Optional[ModuleSpec]:
        try: return object.__getattribute__(module, __SPEC__)
        except: return None
        
    @classmethod
    def get(cls, module: ModuleType) -> 'LoaderState':
        try: return cls(cls.spec(module).loader_state)
        except: return cls.UNKNOWN

    @classmethod
    def set(cls, module: ModuleType, state: Optional[Union['LoaderState', str]] = None) -> ModuleType:
        try: setattr(cls.spec(module), LOADER_STATE, str(cls(state)))
        except: ...
        return module
        
    @classmethod
    def exec(cls, module: ModuleType) -> ModuleType:
        spec = cls.spec(module)
        with redirect_stdout(io.StringIO()), redirect_stderr(io.StringIO()):
            spec.loader.exec_module(module)
        return cls.set(module, cls.DUCKED)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_type.ipynb.

# %% auto 0
__all__ = ['T', 'O', 'P', 'Ts']

# %% ../nbs/01_type.ipynb 6
#| export


# %% ../nbs/01_type.ipynb 8
from typing import (TypeVar, ParamSpec, TypeVarTuple)

# %% ../nbs/01_type.ipynb 10
#| export

# %% ../nbs/01_type.ipynb 12
#| export

# %% ../nbs/01_type.ipynb 14
#| export

# %% ../nbs/01_type.ipynb 16
T = TypeVar('T')
'''Type variable for generic types''';

O = TypeVar('O')
'''Type variable for generic types reprenesting "other"''';

P = ParamSpec('P')
'''Parameter specification type variable, capturing the type of function parameters.''';

Ts = TypeVarTuple('Ts')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/09_meta.ipynb.

# %% auto 0
__all__ = ['ABCSpecies', 'ABCDuck', 'DuckProtocolMeta', 'DuckModuleMeta']

# %% ../nbs/09_meta.ipynb 6
import sys
from abc import ABCMeta, abstractmethod
from functools import wraps, partial
from inspect import isclass
from importlib import import_module
from importlib.util import (module_from_spec)
from importlib.machinery import ModuleSpec

# %% ../nbs/09_meta.ipynb 8
from types import ModuleType, NoneType, FunctionType
from typing import (Any, Self, Callable, TypeGuard, ClassVar, _ProtocolMeta)

# %% ../nbs/09_meta.ipynb 10
#| export

# %% ../nbs/09_meta.ipynb 12
#| export

# %% ../nbs/09_meta.ipynb 14
from quac.cons import (
    ANNOYING_ATTRS, MODULE_TYPE_ATTRS,
    _DUCK, ALL_ATTRS,
    __BASES__, __PROPS__, __MODULE__, __QUALNAME__, __ANNOTATIONS__,  
    __HUSH__, __DUCK__, __ALTS__, __DUCKMODULE__, __DUCKTYPES__
)
from .type import T, O, P, Ts
from .prot import NotProtocol, OptProtocol, DuckTypeProtocol, DuckSpecProtocol
from .enum import Species, LoaderState
from .errs import DuckException, DuckNameError, DuckImportError, DuckAttributeError, DuckModuleNotFoundError
from .grds import isalias, hasduck, hasload, isclstype, isduck, isduckspec, _istype
from .misc import find_last_spec, prepare_types_for_instancecheck, prepare_ducktypes, prepare_isinstance_types
from .solo import EmptyModule
from .spec import DuckSpec

# %% ../nbs/09_meta.ipynb 17
class ABCParam(ABCMeta):
    __args__: tuple
    _parameterized: bool
    def __getitem__(cls, t: T) -> T: 
        cls.__args__ = tuple(t, )
        cls.__qualname__ = str(cls)
        return cls
    
    def __str__(cls):
        return f'{cls.__name__}[{cls.__args__[0]}]'

# %% ../nbs/09_meta.ipynb 19
class ABCSpecies(ABCMeta):
    __species__: ClassVar[Species] = Species.REG
    '''The specific "species" of the protocol. Defaults to `REG` (regular).'''
        
    def __reg__(cls) -> Self:
        '''Set the "species" to `REG` (regular).'''
        cls.__species__ = Species.REG
        return cls
    
    def __not__(cls) -> Self:
        '''Set the "species" to `NOT` (negated).'''
        cls.__species__ = Species.NOT
        return cls
    
    def __opt__(cls) -> Self:
        '''Set the "species" to `OPT` (optional).'''
        cls.__species__ = Species.OPT
        return cls
    
    def __migrate__(cls, species: Species) -> Self:
        '''Migrate the protocol to a different "species".'''
        match species:
            case Species.REG: return cls.__reg__()
            case Species.NOT: return cls.__not__()
            case Species.OPT: return cls.__opt__()
            case _: return cls # NOTE: fail silently
        return cls
    
    def __hash__(self) -> int:
        return super().__hash__()

# %% ../nbs/09_meta.ipynb 21
class ABCDuck(_ProtocolMeta, ABCSpecies, type):
    __ducktype__: ClassVar[tuple[type[T], ...]] = tuple()
    '''The specific types that the protocol is duck-typed to.''';
    
    __expected__: ClassVar[dict[str, Any]] = dict()
    '''Dictionary of attributes and their expected values to check in the object.''';
    
    __required__: ClassVar[tuple[str, ...]] = tuple()
    '''Iterable of attribute names that the object must have.''';
    
    __allequal__: ClassVar[tuple[str, ...]] = tuple(ALL_ATTRS)
    '''Iterable of attributes to check if their values are all the same in an iterable object.''';

    __extrafns__: ClassVar[dict[str, Callable[[O, T], TypeGuard[T]]]] = dict()
    '''Mapping of attribute names to guard functions for additional checks.''';
    
    __usealias__: ClassVar[bool] = True
    '''Whether or not to use the `isalias` function for instance checking or `__guard__`.''';
    
    
    def __types__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to.'''
        curtypes = list(cls.__ducktype__)
        unducked = prepare_ducktypes(cls)
        cls.__ducktype__ = unducked
        return unducked
    
    def __typesq__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to, including NoneType.'''
        return cls.__types__() + (NoneType, )
    
    def __usetypes__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed and to use.'''
        if cls.__species__ == Species.OPT: return cls.__typesq__()
        return cls.__types__()
    
    def __guard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol.'''
        return isinstance(inst, prepare_isinstance_types(cls.__types__()))
    
    def __guardq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol, including NoneType.'''
        return isinstance(inst, prepare_isinstance_types(cls.__typesq__()))
    
    def __alias__(cls, inst: Any) -> TypeGuard[type[T]]:
        isfn: FunctionType = partial(isalias, 
            attrs = cls.__expected__, 
            hasattrs = cls.__required__, 
            allattrs = cls.__allequal__, 
            mapattrs = cls.__extrafns__
        )
        match cls.__species__:
            case Species.REG: return isfn(inst, types=cls.__types__())
            case Species.NOT: return not isfn(inst, types=cls.__types__())
            case Species.OPT: return isfn(inst, types=cls.__typesq__())
        return isalias(
            inst, 
            types = cls.__types__(), 
            guards = (cls.__guard__, ),
            attrs = cls.__expected__, 
            hasattrs = cls.__required__, 
            allattrs = cls.__allequal__, 
            mapattrs = cls.__extrafns__
        )
    
    def __instancecheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        if cls.__usealias__:  
            return cls.__alias__(inst)
        match cls.__species__:
            case Species.REG: return cls.__guard__(inst)
            case Species.NOT: return cls.__notinstance__(inst)
            case Species.OPT: return cls.__optinstance__(inst)
            case _: return cls.__guard__(inst)
        return cls.__guard__(inst)
    
    def __subclasscheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        # print('subclass check', cls, type(cls), inst, type(inst), )
        match cls.__species__:
            case Species.REG: return issubclass(inst, cls.__types__())
            case Species.NOT: return not issubclass(inst, cls.__types__())
            case Species.OPT: return issubclass(inst, cls.__typesq__())
        return issubclass(inst, cls.__types__())
    
    def __notinstance__(cls, inst: Any) -> TypeGuard[NotProtocol[T]]:
        '''Return a type guard for NOT the protocol.'''
        return not cls.__instancecheck__(inst)

    def __optinstance__(cls, inst: Any) -> TypeGuard[OptProtocol[T]]:
        '''Return a type guard for the protocol or NoneType.'''
        return cls.__guardq__(inst)

    def __invert__(cls) -> NotProtocol[T]:
        '''Return the NOT type guard for the protocol.'''
        return cls.__notinstance__

    def __memberguard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        types = cls.__types__()
        if cls.__species__ == Species.OPT: types = cls.__typesq__()
        def checkmember(inst: Any) -> TypeGuard[type[T]]:
            return bool((inst in types) ^ (cls.__species__ == Species.NOT))
        return checkmember

    def __membercheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return inst in cls.__types__()
            case Species.NOT: return not inst in cls.__types__()
            case Species.OPT: return inst in cls.__typesq__()
            case _: return inst in cls.__types__()

    def __contains__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        return cls.__membercheck__(inst)
    
    def __eq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Equality check for the protocol using custom `__guard__` method'''
        try: 
            if cls.__instancecheck__(inst): 
                return True
        except TypeError as inserror: ...
        try: 
            if cls.__subclasscheck__(inst):
                return True
        except TypeError as suberror: ...
        try: 
            if inst in cls.__types__():
                return True
        except TypeError as inerror: ...
        try: 
            if inst is cls:
                return True
        except TypeError as iserror: ...
        return False
    
    def __repr__(cls) -> TypeGuard[type[T]]:
        name = cls.__name__
        match cls.__species__:
            case Species.REG: return f'<duck {name}>'
            case Species.NOT: return f'<!duck {name}>'
            case Species.OPT: return f'<?duck {name}>'
            
    def __call__(cls, *args, **kwargs) -> TypeGuard[type[T]]:
        '''Return the protocol class itself.'''

        fn = cls.__instancecheck__
            
        
        # print(cls, args, fn, fn.__name__)
        results = [fn(arg) for arg in args]
        if len(results) == 1: return results[0]
        return tuple(results)

    def __hash__(self) -> int:
        return super().__hash__()
    
    def __new__(mcls, name, bases, namespace, **kwargs):
        '''Create a new protocol class.'''
        # print('meta.__new__\t', mcls, name, bases, namespace, kwargs)
        new = super().__new__(mcls, name, bases, namespace)
        new = mcls.__metahook__(new, name, bases, namespace, **kwargs)
        return new
        
    @classmethod
    def __metahook__(mcls, cls, name, bases, namespace, **kwargs):
        '''Hook for the protocol metaclass.'''
        # print('meta.__metahook__\t', mcls, cls, name, bases, namespace, kwargs)
        hook = getattr(cls, '__metahook__', None)
        if callable(hook): 
            adj = hook(mcls, name, bases, namespace, **kwargs)
        return cls
    
    def __init__(cls, *args, **kargs):
        ...
        # print('__init__\t', cls, args, kargs)
    
    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__()
        # print('meta.__init_subclass__\t', cls, args, kwargs)
        return 

# %% ../nbs/09_meta.ipynb 36
class DuckProtocolMeta(_ProtocolMeta, ABCSpecies, ABCMeta, type):
    __ducktype__: ClassVar[tuple[type[T], ...]] = tuple()
    '''The specific types that the protocol is duck-typed to.''';
    
    _duck: ClassVar[tuple[type[T], ...]]
    @property
    def __duck__(cls) -> tuple[type[T], ...]:
        if getattr(cls, _DUCK, None) is None:
            unducked = prepare_types_for_instancecheck(cls.__ducktype__)
            setattr(cls, _DUCK, unducked)
        return cls._duck
    
    @abstractmethod
    def __types__(cls) -> tuple[type, ...]:
        '''Return the types that the protocol is duck-typed to.'''
        return prepare_types_for_instancecheck(cls.__duck__)
        duck = list(cls._duck)
        for i, d in enumerate(duck):
            if not (isinstance(d, type) or isclass(d)):
                duck[i] = type(d)
        return tuple(duck)

    @abstractmethod
    def __typesq__(cls) -> tuple[type, ...]: 
        '''Return the types that the protocol is duck-typed to, including NoneType.'''
        return cls.__types__() + (NoneType, )
        
    @abstractmethod
    def __guard__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol.'''
        return isinstance(inst, cls.__types__())
    
    @abstractmethod
    def __guardq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Return a type guard for the protocol, including NoneType.'''
        return isinstance(inst, cls.__typesq__())
        
    def __instancecheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return cls.__guard__(inst)
            case Species.NOT: return cls.__notinstance__(inst)
            case Species.OPT: return cls.__optinstance__(inst)
        return cls.__guard__(inst)
    
    def __subclasscheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return issubclass(inst, cls.__types__())
            case Species.NOT: return not issubclass(inst, cls.__types__())
            case Species.OPT: return issubclass(inst, cls.__typesq__())
        return issubclass(inst, cls.__types__())
    
    @abstractmethod
    def __notinstance__(cls, inst: Any) -> TypeGuard[NotProtocol[T]]:
        '''Return a type guard for NOT the protocol.'''
        return not cls.__instancecheck__(inst)
    
    @abstractmethod
    def __optinstance__(cls, inst: Any) -> TypeGuard[OptProtocol[T]]:
        '''Return a type guard for the protocol or NoneType.'''
        return cls.__guardq__(inst)
    
    @abstractmethod
    def __invert__(cls) -> NotProtocol[T]:
        '''Return the NOT type guard for the protocol.'''
        return cls.__notinstance__
    
    @abstractmethod
    def __membercheck__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Member check for the protocol using custom `__types__` method
        which enables duck-typing and inheritance.'''
        match cls.__species__:
            case Species.REG: return inst in cls.__types__()
            case Species.NOT: return not inst in cls.__types__()
            case Species.OPT: return inst in cls.__typesq__()
        return inst in cls.__types__()
    
    def __contains__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Instance check for the protocol using custom `__guard__` method
        which enables duck-typing and inheritance.'''
        return cls.__membercheck__(inst)
    
    @abstractmethod
    def __eq__(cls, inst: Any) -> TypeGuard[type[T]]: 
        '''Equality check for the protocol using custom `__guard__` method'''
        try: 
            if cls.__instancecheck__(inst): 
                return True
        except TypeError as inserror: ...
        try: 
            if cls.__subclasscheck__(inst):
                return True
        except TypeError as suberror: ...
        try: 
            if inst in cls.__types__():
                return True
        except TypeError as inerror: ...
        try: 
            if inst is cls:
                return True
        except TypeError as iserror: ...
        return False
    
    def __hash__(self) -> int:
        return super().__hash__()

# %% ../nbs/09_meta.ipynb 38
class DuckModuleMeta(ABCMeta, type):
    __duckmodule__: ClassVar[str]
    '''The name of the module to duck.''';
    
    __hush__: ClassVar[bool] = True
    '''Whether or not to raise AttributeError when an attribute is not found.''';
    
    __alts__: ClassVar[dict[str, type]] = {}
    '''Alternatives to use when an attribute is not found.''';
    
    def __instancecheck__(cls: type[T], obj: Any) -> bool:
        '''Check if the object is an instance of the class.'''
        issubbed = isinstance(obj, (ModuleType, ))
        samename = object.__getattribute__(cls, __DUCKMODULE__) == obj.__name__
        hasattrs = all(hasattr(obj, attr) for attr in MODULE_TYPE_ATTRS)
        hasprivs = all(hasattr(obj, attr) for attr in (__DUCKMODULE__, __HUSH__, __ALTS__))
        if issubbed and samename and (hasattrs or hasprivs): return True
        return False
     
    
    def __getattr__(cls: type[T], name: str) -> Any:
        hush = object.__getattribute__(cls, __HUSH__)
        try:
            if name in ANNOYING_ATTRS: 
                try:
                    # return super().__getattribute__(name)
                    raise AttributeError(f"Module '{cls.__name__}' has no attribute '{name}'.")
                except:
                    if object.__getattribute__(cls, __HUSH__): return EmptyModule
                    raise AttributeError(f'''Module '{cls.__name__}' has no attribute '{name}'.''')
            
            # base module name
            base = object.__getattribute__(cls, __DUCKMODULE__)
            # full module name
            full = f'{base}.{name}' if cls.__name__ != 'DuckModule' else name
            
            mkey = full # key to use for sys.modules
            spec: ModuleSpec | None = None # module spec
            stub: str | None = None
            
            # <base.name> not in sys modules
            if full not in sys.modules:            
                # try using importlib to load the module
                try:
                    module = import_module(full)
                    sys.modules[full] = module
                    mkey = full
                except ModuleNotFoundError: ...
                    
                # try using the last found module spec
                try:
                    spec, stub = find_last_spec(full)
                    module: ModuleType = module_from_spec(spec)
                    sys.modules[stub] = module
                    mkey = stub
                    
                except:
                    if object.__getattribute__(cls, __HUSH__): return EmptyModule
                    raise AttributeError(f'''Module '{cls.__name__}' has no attribute '{name}'.''')
                
            # <base.name> or <stub> in sys.modules
            module = sys.modules[mkey]
            
            # create a DuckModule subclass for the submodule if needed            
            # @wraps(cls, assigned=('__name__', '__qualname__', '__doc__', '__annotations__'), updated=())
            # class ducked(cls):
            #     __duckmodule__ = mkey
                
            ducked = type(mkey, (cls, ), dict(__duckmodule__=mkey, __alts__=cls.__alts__, __hush__=cls.__hush__,))
            # ducked = ducked(module.__name__, module.__doc__)
            # ducked.__dict__ = {**module.__dict__, **ducked.__dict__}
            # ducked.__dict__.update({ **module.__dict__, **ducked.__dict__,**dict(_duck=mkey, __alts__=cls.__alts__, __hush__=cls.__hush__,)})

            # Try to get the attribute from the module if it exists
            try: 
                attr = object.__getattribute__(module, name)
                if LoaderState.get(module) not in [LoaderState.NONE, LoaderState.DUCKED]:
                    sys.modules[mkey] = LoaderState.set(module, LoaderState.EXECED)
                return attr
            except AttributeError: ...
            
            # Try actually loading the module if it hasn't been loaded yet
            try: 
                sys.modules[mkey] = LoaderState.exec(module)
                return object.__getattribute__(module, name)
            except AttributeError: ...
            
            # Handle fallbacks if defined
            if hasattr(cls, __ALTS__) and name in cls.__alts__: return cls.__alts__[name]
            
            return ducked
        
        except NameError as e:
            if not hush: raise DuckNameError(*e.args)
        
        except ModuleNotFoundError as e:
            if not hush: raise DuckModuleNotFoundError(e.args, e.name, e.path)
        
        except ImportError as e:
            if not hush: raise DuckImportError(e.args, e.name, e.path)
        
        except AttributeError as e:
            if not hush: raise DuckAttributeError(e.args, e.name, e.obj)
        
        except Exception as e:
            if not hush: raise DuckException(f'NameError: {e}')
        
        return EmptyModule
    
    def __hash__(self) -> int:
        return super().__hash__()
